@@ -11,16 +11,15 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+# You should have received a copy of the GNU General Public License along
+# with this program; if not, write to the Free Software Foundation, Inc.,
+# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 """
 Handle http links.
 """
 
 import urlparse
 import urllib
-import time
 import re
 import zlib
 import socket
@@ -32,7 +31,7 @@ from .. import (log, LOG_CHECK, gzip2 as gzip, strformat, url as urlutil,
 from . import (internpaturl, proxysupport, httpheaders as headers, urlbase,
     get_url_from)
 # import warnings
-from .const import WARN_HTTP_ROBOTS_DENIED, WARN_HTTP_NO_ANCHOR_SUPPORT, \
+from .const import WARN_HTTP_ROBOTS_DENIED, \
     WARN_HTTP_WRONG_REDIRECT, WARN_HTTP_MOVED_PERMANENT, \
     WARN_HTTP_EMPTY_CONTENT, WARN_HTTP_COOKIE_STORE_ERROR, \
     WARN_HTTP_DECOMPRESS_ERROR, WARN_HTTP_UNSUPPORTED_ENCODING, \
@@ -111,9 +110,6 @@ class HttpUrl (internpaturl.InternPatternUrl, proxysupport.ProxySupport):
         super(HttpUrl, self).reset()
         self.max_redirects = 5
         self.has301status = False
-        # some servers do not support anchors in requests
-        # this flag tells us to remove the anchor in request url
-        self.no_anchor = False
         # flag if check had to fallback from HEAD to GET method
         self.fallback_get = False
         # flag if connection is persistent
@@ -124,6 +120,10 @@ class HttpUrl (internpaturl.InternPatternUrl, proxysupport.ProxySupport):
         self.headers = None
         self.auth = None
         self.cookies = []
+        # temporary data filled when reading redirections
+        self._data = None
+        # flag indicating connection reuse
+        self.reused_connection = False
 
     def allows_robots (self, url):
         """
@@ -140,7 +140,7 @@ class HttpUrl (internpaturl.InternPatternUrl, proxysupport.ProxySupport):
         rb = self.aggregate.robots_txt
         callback = self.aggregate.connections.host_wait
         return rb.allows_url(roboturl, url, self.proxy, user, password,
-            callback=callback,)
+            callback=callback)
 
     def check_connection (self):
         """
@@ -187,10 +187,6 @@ class HttpUrl (internpaturl.InternPatternUrl, proxysupport.ProxySupport):
             self.add_info(_("Server `%(name)s' did not support HEAD request; "
                             "a GET request was used instead.") %
                             {"name": server})
-        if self.no_anchor:
-            self.add_warning(_("Server `%(name)s' had no anchor support, removed"
-                               " anchor from request.") % {"name": server},
-                             tag=WARN_HTTP_NO_ANCHOR_SUPPORT)
         # redirections might have changed the URL
         newurl = urlparse.urlunsplit(self.urlparts)
         if self.url != newurl:
@@ -217,7 +213,7 @@ Use URL `%(newurl)s' instead for checking.""") % {
             if response is not None:
                 response.close()
             try:
-                response = self._get_http_response()
+                response = self._try_http_response()
             except httplib.BadStatusLine:
                 # some servers send empty HEAD replies
                 if self.method == "HEAD":
@@ -245,7 +241,7 @@ Use URL `%(newurl)s' instead for checking.""") % {
                          valid=False)
                     return response
                 response.close()
-                response = self._get_http_response()
+                response = self._try_http_response()
                 # restore old proxy settings
                 self.proxy, self.proxyauth = oldproxy
             try:
@@ -283,9 +279,6 @@ Use URL `%(newurl)s' instead for checking.""") % {
                         "Authentication %s/%s", _user, _password)
                     continue
             elif response.status >= 400:
-                if self.headers and self.urlparts[4] and not self.no_anchor:
-                    self.no_anchor = True
-                    continue
                 # retry with GET (but do not set fallback flag)
                 if self.method == "HEAD":
                     self.method = "GET"
@@ -393,7 +386,7 @@ Use URL `%(newurl)s' instead for checking.""") % {
                 return -1, response
             # new response data
             response.close()
-            response = self._get_http_response()
+            response = self._try_http_response()
             tries += 1
         return tries, response
 
@@ -405,13 +398,11 @@ Use URL `%(newurl)s' instead for checking.""") % {
         data = self.get_cache_data()
         data["warnings"] = [
             x for x in self.warnings if x[0] != "http-moved-permanent"]
-        data["info"] = [x for x in self.info if x[0] != "http-redirect"]
+        data["info"] = self.info
         return data
 
     def check_response (self, response):
-        """
-        Check final result and log it.
-        """
+        """Check final result and log it."""
         if response.status >= 400:
             self.set_result(u"%r %s" % (response.status, response.reason),
                             valid=False)
@@ -444,6 +435,21 @@ Use URL `%(newurl)s' instead for checking.""") % {
         if modified:
             self.add_info(_("Last modified %(date)s.") % {"date": modified})
 
+    def _try_http_response (self):
+        """Try to get a HTTP response object. For reused persistent
+        connections that the server closed unexpected, a new connection
+        will be opened.
+        """
+        try:
+            return self._get_http_response()
+        except socket.error, msg:
+            if msg.args[0] == 32 and self.reused_connection:
+                # server closed persistent connection - retry
+                self.persistent = False
+                self.close_connection()
+                return self._get_http_response()
+            raise
+
     def _get_http_response (self):
         """
         Send HTTP request and get response object.
@@ -458,10 +464,9 @@ Use URL `%(newurl)s' instead for checking.""") % {
         # close/release a previous connection
         self.close_connection()
         self.url_connection = self.get_http_object(host, scheme)
-        if self.no_anchor:
-            anchor = ''
-        else:
-            anchor = self.urlparts[4]
+        # the anchor fragment is not part of a HTTP URL, see
+        # http://tools.ietf.org/html/rfc2616#section-3.2.2
+        anchor = ''
         if self.proxy:
             path = urlparse.urlunsplit((self.urlparts[0], self.urlparts[1],
                                  self.urlparts[2], self.urlparts[3], anchor))
@@ -513,6 +518,7 @@ Use URL `%(newurl)s' instead for checking.""") % {
            self.headers.getheader("Content-Length") != "0"):
             # always read content from persistent connections
             self._read_content(response)
+            assert not response.will_close
         # If possible, use official W3C HTTP response name
         if response.status in httpresponses:
             response.reason = httpresponses[response.status]
@@ -534,6 +540,7 @@ Use URL `%(newurl)s' instead for checking.""") % {
         conn = self.aggregate.connections.get(key)
         if conn is not None:
             log.debug(LOG_CHECK, "reuse cached HTTP(S) connection %s", conn)
+            self.reused_connection = True
             return conn
         self.aggregate.connections.wait_for_host(host)
         if scheme == "http":
@@ -548,26 +555,24 @@ Use URL `%(newurl)s' instead for checking.""") % {
         h.connect()
         return h
 
-    def get_content (self):
-        """
-        Get content of the URL target. The content data is cached after
+    def read_content (self):
+        """Get content of the URL target. The content data is cached after
         the first call to this method.
 
         @return: URL content, decompressed and decoded
         @rtype: string
         """
-        if self.data is None:
-            self.method = "GET"
-            response = self._get_http_response()
-            response = self.follow_redirections(response, set_result=False)[1]
-            self.headers = response.msg
+        self.method = "GET"
+        response = self._try_http_response()
+        response = self.follow_redirections(response, set_result=False)[1]
+        self.headers = response.msg
+        if self._data is None:
             self._read_content(response)
-            if self.data is None:
-                self.data = ""
-        return self.data
+        data = self._data
+        self._data = None
+        return data
 
     def _read_content (self, response):
-        t = time.time()
         data = response.read()
         encoding = headers.get_content_encoding(self.headers)
         if encoding in _supported_encodings:
@@ -581,11 +586,12 @@ Use URL `%(newurl)s' instead for checking.""") % {
                                  {"err": str(msg)},
                                  tag=WARN_HTTP_DECOMPRESS_ERROR)
                 f = StringIO(data)
-            data = f.read()
-        if self.data is None and self.method == "GET" and \
-           response.status not in [301, 302]:
-            self.data = data
-            self.dltime = time.time() - t
+            try:
+                data = f.read()
+            finally:
+                f.close()
+        # store temporary data
+        self._data = data
 
     def encoding_supported (self):
         """Check if page encoding is supported."""
